/*******************************************************************************
 * Copyright (c) 2005 Matthew Hall and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Matthew Hall - initial API and implementation
 ******************************************************************************/
package formslayout;

import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;

/**
 * Helper class for laying out controls on a composites using a FormLayout.
 * 
 * @author Matthew
 */
public class FormBuilder {
	private final Composite composite;
	private final FormLayout layout;

	/**
	 * Constructs a FormBuilder for building a form on the given Composite,
	 * using the given FormLayout.
	 * 
	 * @param composite
	 * 		the composite on which the form is being built.
	 * @param layout
	 * 		the FormLayout being used.
	 */
	public FormBuilder(Composite composite, FormLayout layout) {
		if (composite == null || layout == null)
			throw new NullPointerException();

		this.composite = composite;
		this.layout = layout;
		composite.setLayout(layout);
	}

	/**
	 * Returns the composite on which the form is being built.
	 * 
	 * @return the composite on which the form is being built.
	 */
	public Composite getComposite() {
		return composite;
	}

	/**
	 * Returns the FormLayout being used on the composite.
	 * 
	 * @return the FormLayout being used on the composite.
	 */
	public FormLayout getLayout() {
		return layout;
	}

	/* checkFormData has already been called before this method is invoked. */
	private <T extends Control> T _add(T control, FormData data) {
		if (control.getParent() != composite)
			throw new IllegalArgumentException("Control " + control
					+ " does not belong to this FormBuilder's composite");
		control.setLayoutData(data);
		return control;
	}

	/**
	 * Adds the given control at (x, y) in the form.
	 * 
	 * @param <
	 * 		T> the type of control.
	 * @param control
	 * 		the control to add.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @return the control that was added.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public <T extends Control> T add(T control, int x, int y) {
		return add(control, FormData.xy(x, y));
	}

	/**
	 * Adds the given control at (x, y) in the form.
	 * 
	 * @param <
	 * 		T> the type of control.
	 * @param control
	 * 		the control to add.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the control that was added.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public <T extends Control> T add(T control, int x, int y, String align) {
		return add(control, FormData.xy(x, y, align));
	}

	/**
	 * Adds the given control at the given cellspan in the form.
	 * 
	 * @param <
	 * 		T> the type of control.
	 * @param control
	 * 		the control to add.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @return the control that was added.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public <T extends Control> T add(T control, int x, int y, int w) {
		return add(control, FormData.xyw(x, y, w));
	}

	/**
	 * Adds the given control at the given cellspan in the form.
	 * 
	 * @param <
	 * 		T> the type of control.
	 * @param control
	 * 		the control to add.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the control that was added.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public <T extends Control> T add(T control, int x, int y, int w,
			String align) {
		return add(control, FormData.xyw(x, y, w, align));
	}

	/**
	 * Adds the given control at the given cellspan in the form.
	 * 
	 * @param <
	 * 		T> the type of control.
	 * @param control
	 * 		the control to add.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @return the control that was added.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public <T extends Control> T add(T control, int x, int y, int w, int h) {
		return add(control, FormData.xywh(x, y, w, h));
	}

	/**
	 * Adds the given control at the given cellspan in the form.
	 * 
	 * @param <
	 * 		T> the type of control.
	 * @param control
	 * 		the control to add.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the control that was added.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public <T extends Control> T add(T control, int x, int y, int w, int h,
			String align) {
		return add(control, FormData.xywh(x, y, w, h, align));
	}

	/**
	 * Adds the given control at the cellspan described by the given FormData.
	 * 
	 * @param <
	 * 		T> the type of control.
	 * @param control
	 * 		the control to add.
	 * @param data
	 * 		a FormData describing the cellspan where the control should be added
	 * 		to the form.
	 * @return the control that was added.
	 * @throws IndexOutOfBoundsException
	 * 		if the FormData describes a cell span outside the boundaries of the
	 * 		FormLayout.
	 */
	public <T extends Control> T add(T control, FormData data) {
		layout.checkFormData(data);
		return _add(control, data);
	}

	/**
	 * Adds a new separator label at (x, y) in the form.
	 * 
	 * @param text
	 * 		the new separator label's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @return the separator label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public SeparatorLabel separator(String text, int x, int y) {
		return separator(text, FormData.xy(x, y));
	}

	/**
	 * Adds a new separator label at (x, y) in the form.
	 * 
	 * @param text
	 * 		the new separator label's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the separator label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public SeparatorLabel separator(String text, int x, int y, String align) {
		return separator(text, FormData.xy(x, y, align));
	}

	/**
	 * Adds a new separator label at the given cellspan in the form.
	 * 
	 * @param text
	 * 		the new separator label's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @return the separator label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public SeparatorLabel separator(String text, int x, int y, int w) {
		return separator(text, FormData.xyw(x, y, w));
	}

	/**
	 * Adds a new separator label at the given cellspan in the form.
	 * 
	 * @param text
	 * 		the new separator label's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the separator label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public SeparatorLabel separator(String text, int x, int y, int w,
			String align) {
		return separator(text, FormData.xyw(x, y, w, align));
	}

	/**
	 * Adds a new separator label at the given cellspan in the form.
	 * 
	 * @param text
	 * 		the new separator label's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @return the separator label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public SeparatorLabel separator(String text, int x, int y, int w, int h) {
		return separator(text, FormData.xywh(x, y, w, h));
	}

	/**
	 * Adds a new separator label at the given cellspan in the form.
	 * 
	 * @param text
	 * 		the new separator label's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the separator label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public SeparatorLabel separator(String text, int x, int y, int w, int h,
			String align) {
		return separator(text, FormData.xywh(x, y, w, h, align));
	}

	/**
	 * Adds a new separator label at the cellspan described by the given
	 * FormData.
	 * 
	 * @param text
	 * 		the new separator label's text property.
	 * @param data
	 * 		a FormData describing the cellspan where the label should be added
	 * 		to the form.
	 * @return the separator label that was created.
	 * @throws IndexOutOfBoundsException
	 * 		if the FormData describes a cell span outside the boundaries of the
	 * 		FormLayout.
	 */
	public SeparatorLabel separator(String text, FormData data) {
		layout.checkFormData(data);
		SeparatorLabel separator = new SeparatorLabel(composite, SWT.NONE);
		separator.setText(text);
		_add(separator, data);
		return separator;
	}

	/**
	 * Adds a new label at (x, y) in the form.
	 * 
	 * @param text
	 * 		the new label's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @return the label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Label label(String text, int x, int y) {
		return label(text, FormData.xy(x, y));
	}

	/**
	 * Adds a new label at (x, y) in the form.
	 * 
	 * @param text
	 * 		the new label's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Label label(String text, int x, int y, String align) {
		return label(text, FormData.xy(x, y, align));
	}

	/**
	 * Adds a new label at the given cellspan in the form.
	 * 
	 * @param text
	 * 		the new label's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @return the label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Label label(String text, int x, int y, int w) {
		return label(text, FormData.xyw(x, y, w));
	}

	/**
	 * Adds a new label at the given cellspan in the form.
	 * 
	 * @param text
	 * 		the new label's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Label label(String text, int x, int y, int w, String align) {
		return label(text, FormData.xyw(x, y, w, align));
	}

	/**
	 * Adds a new label at the given cellspan in the form.
	 * 
	 * @param text
	 * 		the new label's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @return the label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Label label(String text, int x, int y, int w, int h) {
		return label(text, FormData.xywh(x, y, w, h));
	}

	/**
	 * Adds a new label at the given cellspan in the form.
	 * 
	 * @param text
	 * 		the new label's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Label label(String text, int x, int y, int w, int h, String align) {
		return label(text, FormData.xywh(x, y, w, h, align));
	}

	/**
	 * Adds a new label at the cellspan described by the given FormData.
	 * 
	 * @param text
	 * 		the new label's text property.
	 * @param data
	 * 		a FormData describing the cellspan where the label should be added
	 * 		to the form.
	 * @return the label that was created.
	 * @throws IndexOutOfBoundsException
	 * 		if the FormData describes a cell span outside the boundaries of the
	 * 		FormLayout.
	 */
	public Label label(String text, FormData data) {
		layout.checkFormData(data);
		Label label = new Label(composite, SWT.NONE);
		label.setText(text);
		return _add(label, data);
	}

	/**
	 * Adds a new button at (x, y) in the form.
	 * 
	 * @param style
	 * 		the button style.
	 * @param text
	 * 		the new button's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @return the button that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Button button(int style, String text, int x, int y) {
		return button(style, text, FormData.xy(x, y));
	}

	/**
	 * Adds a new button at (x, y) in the form.
	 * 
	 * @param style
	 * 		the button style.
	 * @param text
	 * 		the new button's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the button that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Button button(int style, String text, int x, int y, String align) {
		return button(style, text, FormData.xy(x, y, align));
	}

	/**
	 * Adds a new button at the given cellspan in the form.
	 * 
	 * @param style
	 * 		the button style.
	 * @param text
	 * 		the new button's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @return the button that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Button button(int style, String text, int x, int y, int w) {
		return button(style, text, FormData.xyw(x, y, w));
	}

	/**
	 * Adds a new button at the given cellspan in the form.
	 * 
	 * @param style
	 * 		the button style.
	 * @param text
	 * 		the new button's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the button that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Button button(int style, String text, int x, int y, int w,
			String align) {
		return button(style, text, FormData.xyw(x, y, w, align));
	}

	/**
	 * Adds a new button at the given cellspan in the form.
	 * 
	 * @param style
	 * 		the button style.
	 * @param text
	 * 		the new button's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @return the button that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Button button(int style, String text, int x, int y, int w, int h) {
		return button(style, text, FormData.xywh(x, y, w, h));
	}

	/**
	 * Adds a new button at the given cellspan in the form.
	 * 
	 * @param style
	 * 		the button style.
	 * @param text
	 * 		the new button's text property.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the button that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Button button(int style, String text, int x, int y, int w, int h,
			String align) {
		return button(style, text, FormData.xywh(x, y, w, h, align));
	}

	/**
	 * Adds a new button at the cellspan described by the given FormData.
	 * 
	 * @param style
	 * 		the button style.
	 * @param text
	 * 		the new button's text property.
	 * @param data
	 * 		a FormData describing the cellspan where the label should be added
	 * 		to the form.
	 * @return the button that was created.
	 * @throws IndexOutOfBoundsException
	 * 		if the FormData describes a cell span outside the boundaries of the
	 * 		FormLayout.
	 */
	public Button button(int style, String text, FormData data) {
		layout.checkFormData(data);
		Button button = new Button(composite, style);
		button.setText(text);
		return _add(button, data);
	}

	/**
	 * Adds a new single-line text at (x, y) in the form.
	 * 
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @return the text that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Text text(int x, int y) {
		return text(FormData.xy(x, y));
	}

	/**
	 * Adds a new single-line text at (x, y) in the form.
	 * 
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the text that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Text text(int x, int y, String align) {
		return text(FormData.xy(x, y, align));
	}

	/**
	 * Adds a new single-line text at the given cellspan in the form.
	 * 
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @return the text that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Text text(int x, int y, int w) {
		return text(FormData.xyw(x, y, w));
	}

	/**
	 * Adds a new single-line text at the given cellspan in the form.
	 * 
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the text that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Text text(int x, int y, int w, String align) {
		return text(FormData.xyw(x, y, w));
	}

	/**
	 * Adds a new single-line text at the given cellspan in the form.
	 * 
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @return the text that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Text text(int x, int y, int w, int h) {
		return text(FormData.xywh(x, y, w, h));
	}

	/**
	 * Adds a new single-line text at the given cellspan in the form.
	 * 
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the text that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Text text(int x, int y, int w, int h, String align) {
		return text(FormData.xywh(x, y, w, h, align));
	}

	/**
	 * Adds a new single-line text at the cellspan described by the given
	 * FormData.
	 * 
	 * @param data
	 * 		a FormData describing the cellspan where the label should be added
	 * 		to the form.
	 * @return the text that was created.
	 * @throws IndexOutOfBoundsException
	 * 		if the FormData describes a cell span outside the boundaries of the
	 * 		FormLayout.
	 */
	public Text text(FormData data) {
		layout.checkFormData(data);
		return _add(new Text(composite, SWT.SINGLE | SWT.BORDER), data);
	}

	/**
	 * Adds a new multi-line text at (x, y) in the form.
	 * 
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @return the text that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Text textArea(int x, int y) {
		return textArea(FormData.xy(x, y));
	}

	/**
	 * Adds a new multi-line text at (x, y) in the form.
	 * 
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the text that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Text textArea(int x, int y, String align) {
		return textArea(FormData.xy(x, y, align));
	}

	/**
	 * Adds a new multi-line text at the given cellspan in the form.
	 * 
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @return the text that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Text textArea(int x, int y, int w) {
		return textArea(FormData.xyw(x, y, w));
	}

	/**
	 * Adds a new multi-line text at the given cellspan in the form.
	 * 
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the text that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Text textArea(int x, int y, int w, String align) {
		return textArea(FormData.xyw(x, y, w, align));
	}

	/**
	 * Adds a new multi-line text at the given cellspan in the form.
	 * 
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @return the text that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Text textArea(int x, int y, int w, int h) {
		return textArea(FormData.xywh(x, y, w, h));
	}

	/**
	 * Adds a new multi-line text at the given cellspan in the form.
	 * 
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the text that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Text textArea(int x, int y, int w, int h, String align) {
		return textArea(FormData.xywh(x, y, w, h, align));
	}

	/**
	 * Adds a new multi-line text at the cellspan described by the given
	 * FormData.
	 * 
	 * @param data
	 * 		a FormData describing the cellspan where the label should be added
	 * 		to the form.
	 * @return the text that was created.
	 * @throws IndexOutOfBoundsException
	 * 		if the FormData describes a cell span outside the boundaries of the
	 * 		FormLayout.
	 */
	public Text textArea(FormData data) {
		layout.checkFormData(data);
		return _add(new Text(composite, SWT.MULTI | SWT.BORDER | SWT.WRAP
				| SWT.V_SCROLL), data);
	}

	/**
	 * Adds a new combo at (x, y) in the form.
	 * 
	 * @param style
	 * 		the combo style.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @return the combo that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Combo combo(int style, int x, int y) {
		return combo(style, FormData.xy(x, y));
	}

	/**
	 * Adds a new combo at (x, y) in the form.
	 * 
	 * @param style
	 * 		the combo style.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the combo that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Combo combo(int style, int x, int y, String align) {
		return combo(style, FormData.xy(x, y, align));
	}

	/**
	 * Adds a new combo at the given cellspan in the form.
	 * 
	 * @param style
	 * 		the combo style.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @return the label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Combo combo(int style, int x, int y, int w) {
		return combo(style, FormData.xyw(x, y, w));
	}

	/**
	 * Adds a new combo at the given cellspan in the form.
	 * 
	 * @param style
	 * 		the combo style.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Combo combo(int style, int x, int y, int w, String align) {
		return combo(style, FormData.xyw(x, y, w, align));
	}

	/**
	 * Adds a new combo at the given cellspan in the form.
	 * 
	 * @param style
	 * 		the combo style.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @return the label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Combo combo(int style, int x, int y, int w, int h) {
		return combo(style, FormData.xywh(x, y, w, h));
	}

	/**
	 * Adds a new combo at the given cellspan in the form.
	 * 
	 * @param style
	 * 		the combo style.
	 * @param x
	 * 		the column index of the cell where the control will be added.
	 * @param y
	 * 		the row index of the cell where the control will be added.
	 * @param w
	 * 		the horizontal cell span.
	 * @param h
	 * 		the vertical cell span.
	 * @param align
	 * 		a comma-delimited string containing a parseable horizontal alignment
	 * 		following by a parseable vertical alignment. See the
	 * 		ColSpec.parse(String) and RowSpec.parse(String) methods for the
	 * 		possible alignment values.
	 * @return the label that was created.
	 * @throws IllegalArgumentException
	 * 		if any of the arguments are invalid.
	 * @throws IndexOutOfBoundsException
	 * 		if any of the row or column indices are out of bounds.
	 */
	public Combo combo(int style, int x, int y, int w, int h, String align) {
		return combo(style, FormData.xywh(x, y, w, h, align));
	}

	/**
	 * Adds a new combo at the cellspan described by the given FormData.
	 * 
	 * @param style
	 * 		the combo style.
	 * @param data
	 * 		a FormData describing the cellspan where the label should be added
	 * 		to the form.
	 * @return the label that was created.
	 * @throws IndexOutOfBoundsException
	 * 		if the FormData describes a cell span outside the boundaries of the
	 * 		FormLayout.
	 */
	public Combo combo(int style, FormData data) {
		layout.checkFormData(data);
		return _add(new Combo(composite, style), data);
	}
}