/*******************************************************************************
 * Copyright (c) 2006 Matthew Hall and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     Matthew Hall - initial API and implementation
 ******************************************************************************/
package formslayout;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontMetrics;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.widgets.Display;

/**
 * Abstract base class of ColSpec and RowSpec.
 * 
 * @author Matthew
 */
public abstract class FormSpec {
	static final int DEFAULT_SIZE = SWT.DEFAULT;
	static final int DEFAULT_WEIGHT = 0;

	/** Alignment field. Value depends on subclass. */
	public final int align;
	/** Sizes field. Value may be SWT.DEFAULT, FormLayout.PREFERRED, or >= 0 */
	public final int size;
	/** Weight field. Value may be >= 0. */
	public final int weight;

	FormSpec(int align, int size, int weight) {
		this.align = checkAlign(align);
		this.size = checkSize(size);
		this.weight = checkWeight(weight);
	}

	abstract int checkAlign(int align);

	private int checkSize(int size) {
		if (size == SWT.DEFAULT || size == FormLayout.PREFERRED || size >= 0)
			return size;

		throw new IllegalArgumentException("Illegal size value: " + size);
	}

	private int checkWeight(int weight) {
		if (weight < 0)
			throw new IllegalArgumentException("Illegal weight value: "
					+ weight + " (must be non-negative)");

		return weight;
	}

	// Size patterns.
	private static final Pattern PREFERRED_SIZE_PATTERN = Pattern.compile(
			"p|pref|preferred", Pattern.CASE_INSENSITIVE);
	private static final Pattern DEFAULT_SIZE_PATTERN = Pattern.compile(
			"d|def|default", Pattern.CASE_INSENSITIVE);
	private static final Pattern DLU_PATTERN = Pattern.compile("(\\d+)dlu",
			Pattern.CASE_INSENSITIVE);
	private static final Pattern PIXEL_PATTERN = Pattern.compile("(\\d+)px",
			Pattern.CASE_INSENSITIVE);
	private static final Pattern ANY_SIZE_PATTERN = Pattern.compile(
			"p|pref|preferred|d|def|default|\\d+(dlu|px)",
			Pattern.CASE_INSENSITIVE);

	static boolean isSize(String size) {
		return ANY_SIZE_PATTERN.matcher(size).matches();
	}

	static int parseSize(String size) {
		Matcher matcher;

		if ((matcher = PREFERRED_SIZE_PATTERN.matcher(size)).matches())
			return FormLayout.PREFERRED;
		if ((matcher = DEFAULT_SIZE_PATTERN.matcher(size)).matches())
			return SWT.DEFAULT;
		else if ((matcher = DLU_PATTERN.matcher(size)).matches()) {
			int DLUs = Integer.parseInt(matcher.group(1));
			Font font = Display.getDefault().getSystemFont();
			GC gc = new GC(Display.getDefault());
			try {
				gc.setFont(font);
				FontMetrics fm = gc.getFontMetrics();
				return DLUs * fm.getHeight() / 8;
			} finally {
				gc.dispose();
			}
		} else if ((matcher = PIXEL_PATTERN.matcher(size)).matches()) {
			return Integer.parseInt(matcher.group(1));
		} else {
			throw new IllegalArgumentException("Unknown size pattern: \""
					+ size + "\"");
		}
	}

	private static final Pattern NO_WEIGHT_PATTERN = Pattern.compile("n|none",
			Pattern.CASE_INSENSITIVE);
	private static final Pattern GROW_WEIGHT_PATTERN = Pattern.compile(
			"(g|grow)(\\((\\d+)\\))?", // yikes
			Pattern.CASE_INSENSITIVE);
	private static final Pattern ANY_WEIGHT_PATTERN = Pattern.compile(
			"n|none|(g|grow)(\\(\\d+\\))?", Pattern.CASE_INSENSITIVE);

	static boolean isWeight(String weight) {
		return ANY_WEIGHT_PATTERN.matcher(weight).matches();
	}

	static int parseWeight(String weight) {
		Matcher matcher;
		if (NO_WEIGHT_PATTERN.matcher(weight).matches())
			return 0;
		else if ((matcher = GROW_WEIGHT_PATTERN.matcher(weight)).matches()) {
			return (matcher.group(3) == null) ? 1 : Integer.parseInt(matcher
					.group(3));
		} else
			throw new IllegalArgumentException("Illegal weight pattern: \""
					+ weight + "\"");
	}
}
